@@ -40,6 +40,7 @@
 #include "load_save.h"
 #include "group.h"
 #include "diagramdata.h"
+#include "object.h"
 #include "message.h"
 #include "preferences.h"
 #include "diapagelayout.h"
@@ -60,9 +61,9 @@ static void read_connections(GList *objects, xmlNodePtr layer_node,
 static void GHFuncUnknownObjects(gpointer key,
 				 gpointer value,
 				 gpointer user_data);
-static GList *read_objects(xmlNodePtr objects, Layer *layer,
+static GList *read_objects(xmlNodePtr objects,
 			   GHashTable *objects_hash,
-			   const char *filename, Object *parent);
+			   const char *filename, DiaObject *parent);
 static void hash_free_string(gpointer       key,
 			     gpointer       value,
 			     gpointer       user_data);
@@ -90,13 +91,30 @@ GHFuncUnknownObjects(gpointer key,
   g_free(key);
 }
 
+/**
+ * Recursive function to read objects from a specific level in the xml.
+ *
+ * Nowadays there are quite a few of them :
+ * - Layers : a diagram may have different layers, but this function does *not*
+ *   add the created objects to them as it does not know on which nesting level it
+ *   is called. So the topmost caller must add the returned objects to the layer.
+ * - Groups : groups in itself can have an arbitrary nesting level including other
+ *   groups or objects or both of them. A group not containing any objects is by 
+ *   definition useless. So it is not created. This is to avoid trouble with some older
+ *   diagrams which happen to be saved with empty groups.
+ * - Parents : if the parent relations would have been there from the beginning of
+ *   Dias file format they probably would have been added as nesting level 
+ *   themselves. But to maintain forward compatibility (allow to let older versions
+ *   of Dia to see as much as possible) they were added all on the same level and
+ *   the parent child relation is reconstructed from additional attributes.
+ */
 static GList *
-read_objects(xmlNodePtr objects, Layer *layer,
-             GHashTable *objects_hash,const char *filename, Object *parent)
+read_objects(xmlNodePtr objects, 
+             GHashTable *objects_hash,const char *filename, DiaObject *parent)
 {
   GList *list;
-  ObjectType *type;
-  Object *obj;
+  DiaObjectType *type;
+  DiaObject *obj;
   ObjectNode obj_node;
   char *typestr;
   char *versionstr;
@@ -112,7 +130,7 @@ read_objects(xmlNodePtr objects, Layer *layer,
   list = NULL;
 
   obj_node = objects->xmlChildrenNode;
-  
+
   while ( obj_node != NULL) {
     if (xmlIsBlankNode(obj_node)) {
       obj_node = obj_node->next;
@@ -141,7 +159,6 @@ read_objects(xmlNodePtr objects, Layer *layer,
       else
       {
         obj = type->ops->load(obj_node, version, filename);
-        layer_add_object(layer,obj);
         list = g_list_append(list, obj);
 
         if (parent)
@@ -158,7 +175,7 @@ read_objects(xmlNodePtr objects, Layer *layer,
         {
           if (strcmp(child_node->name, "children") == 0)
           {
-	    GList *children_read = read_objects(child_node, layer, objects_hash, filename, obj);
+	    GList *children_read = read_objects(child_node, objects_hash, filename, obj);
             list = g_list_concat(list, children_read);
             break;
           }
@@ -167,11 +184,10 @@ read_objects(xmlNodePtr objects, Layer *layer,
       }
       if (typestr) xmlFree(typestr);
       if (id) xmlFree (id);
-    } else if (strcmp(obj_node->name, "group")==0) {
-      obj = group_create(read_objects(obj_node, layer,
-                                      objects_hash, filename, NULL));
-      layer_add_object(layer,obj);
-      
+    } else if (   strcmp(obj_node->name, "group")==0
+               && obj_node->children) {
+      /* don't create empty groups */
+      obj = group_create(read_objects(obj_node, objects_hash, filename, NULL));
       list = g_list_append(list, obj);
     } else {
       /* silently ignore other nodes */
@@ -184,7 +200,7 @@ read_objects(xmlNodePtr objects, Layer *layer,
     g_hash_table_foreach(unknown_hash,
 			 GHFuncUnknownObjects,
 			 unknown_str);
-    message_error(unknown_str->str);
+    message_error("%s", unknown_str->str);
   }
   g_hash_table_destroy(unknown_hash);
   g_string_free(unknown_str, TRUE);
@@ -204,12 +220,12 @@ read_connections(GList *objects, xmlNodePtr layer_node,
   char *tostr;
   char *connstr;
   int handle, conn;
-  Object *to;
+  DiaObject *to;
   
   list = objects;
   obj_node = layer_node->xmlChildrenNode;
   while ((list != NULL) && (obj_node != NULL)) {
-    Object *obj = (Object *) list->data;
+    DiaObject *obj = (DiaObject *) list->data;
 
     while (obj_node && xmlIsBlankNode(obj_node)) obj_node = obj_node->next;
     if (!obj_node) break;
@@ -224,6 +240,7 @@ read_connections(GList *objects, xmlNodePtr layer_node,
       if (connections != NULL) {
 	connection = connections->xmlChildrenNode;
 	while (connection != NULL) {
+	  char *donestr;
           if (xmlIsBlankNode(connection)) {
             connection = connection->next;
             continue;
@@ -233,28 +250,44 @@ read_connections(GList *objects, xmlNodePtr layer_node,
 	  connstr = xmlGetProp(connection, "connection");
 	  
 	  handle = atoi(handlestr);
-	  conn = atoi(connstr);
-	  
-	  to = g_hash_table_lookup(objects_hash, tostr);
+	  conn = strtol(connstr, &donestr, 10);
+	  if (*donestr != '\0') { /* Didn't convert it all -- use string */
+	    conn = -1;
+	  }
 
-	  if (handlestr) xmlFree(handlestr);
-	  if (connstr) xmlFree(connstr);
-	  if (tostr) xmlFree(tostr);
+	  to = g_hash_table_lookup(objects_hash, tostr);
 
 	  if (to == NULL) {
 	    message_error(_("Error loading diagram.\n"
 			    "Linked object not found in document."));
-	  } else if (conn >= 0 && conn >= to->num_connections) {
-	    message_error(_("Error loading diagram.\n"
-			    "connection point does not exist."));
-	  } else if (handle >= 0 && handle >= obj->num_handles) {
+	  } else if (handle < 0 || handle >= obj->num_handles) {
 	    message_error(_("Error loading diagram.\n"
 			    "connection handle does not exist."));
 	  } else {
-	    object_connect(obj, obj->handles[handle],
-			   to->connections[conn]);
+	    if (conn == -1) { /* Find named connpoint */
+	      int i;
+	      for (i = 0; i < to->num_connections; i++) {
+		if (to->connections[i]->name != NULL &&
+		    strcmp(to->connections[i]->name, connstr) == 0) {
+		  conn = i;
+		  break;
+		}
+	      }
+	    }
+	    if (conn >= 0 && conn < to->num_connections) {
+	      object_connect(obj, obj->handles[handle],
+			     to->connections[conn]);
+	    } else {
+	      message_error(_("Error loading diagram.\n"
+			      "connection point %s does not exist."),
+			    connstr);
+	    }
 	  }
 
+	  if (handlestr) xmlFree(handlestr);
+	  if (tostr) xmlFree(tostr);
+	  if (connstr) xmlFree(connstr);
+
 	  connection = connection->next;
 	}
       }
@@ -313,6 +346,7 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
   Layer *layer;
   xmlNsPtr namespace;
   gchar firstchar;
+  Diagram *diagram = DIA_IS_DIAGRAM (data) ? DIA_DIAGRAM (data) : NULL;
 
   if (g_file_test (filename, G_FILE_TEST_IS_DIR)) {
     message_error(_("You must specify a file, not a directory.\n"));
@@ -322,7 +356,8 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
   fd = open(filename, O_RDONLY);
 
   if (fd==-1) {
-    message_error(_("Couldn't open: '%s' for reading.\n"), filename);
+    message_error(_("Couldn't open: '%s' for reading.\n"),
+		  dia_message_filename(filename));
     return FALSE;
   }
 
@@ -339,19 +374,22 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
   doc = xmlDiaParseFile(filename);
 
   if (doc == NULL){
-    message_error(_("Error loading diagram %s.\nUnknown file type."),filename);
+    message_error(_("Error loading diagram %s.\nUnknown file type."),
+		  dia_message_filename(filename));
     return FALSE;
   }
   
   if (doc->xmlRootNode == NULL) {
-    message_error(_("Error loading diagram %s.\nUnknown file type."),filename);
+    message_error(_("Error loading diagram %s.\nUnknown file type."),
+		  dia_message_filename(filename));
     xmlFreeDoc (doc);
     return FALSE;
   }
 
   namespace = xmlSearchNs(doc, doc->xmlRootNode, "dia");
   if (strcmp (doc->xmlRootNode->name, "diagram") || (namespace == NULL)){
-    message_error(_("Error loading diagram %s.\nNot a Dia file."), filename);
+    message_error(_("Error loading diagram %s.\nNot a Dia file."), 
+		  dia_message_filename(filename));
     xmlFreeDoc (doc);
     return FALSE;
   }
@@ -369,11 +407,12 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
   if (attr != NULL)
     data_color(attribute_first_data(attr), &data->bg_color);
 
-  data->pagebreak_color = prefs.new_diagram.pagebreak_color;
-  attr = composite_find_attribute(diagramdata, "pagebreak");
-  if (attr != NULL)
-    data_color(attribute_first_data(attr), &data->pagebreak_color);
-
+  if (diagram) {
+    diagram->pagebreak_color = prefs.new_diagram.pagebreak_color;
+    attr = composite_find_attribute(diagramdata, "pagebreak");
+    if (attr != NULL)
+      data_color(attribute_first_data(attr), &diagram->pagebreak_color);
+  }
   /* load paper information from diagramdata section */
   attr = composite_find_attribute(diagramdata, "paper");
   if (attr != NULL) {
@@ -384,6 +423,9 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
       g_free(data->paper.name);
       data->paper.name = data_string(attribute_first_data(attr));
     }
+    if (data->paper.name == NULL || data->paper.name[0] == '\0') {
+      data->paper.name = g_strdup(prefs.new_diagram.papertype);
+    }
     /* set default margins for paper size ... */
     dia_page_layout_get_default_margins(data->paper.name,
 					&data->paper.tmargin,
@@ -446,55 +488,59 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
     data->paper.width /= data->paper.scaling;
     data->paper.height /= data->paper.scaling;
   }
-  attr = composite_find_attribute(diagramdata, "grid");
-  if (attr != NULL) {
-    gridinfo = attribute_first_data(attr);
 
-    attr = composite_find_attribute(gridinfo, "width_x");
-    if (attr != NULL)
-      data->grid.width_x = data_real(attribute_first_data(attr));
-    attr = composite_find_attribute(gridinfo, "width_y");
-    if (attr != NULL)
-      data->grid.width_y = data_real(attribute_first_data(attr));
-    attr = composite_find_attribute(gridinfo, "visible_x");
-    if (attr != NULL)
-      data->grid.visible_x = data_int(attribute_first_data(attr));
-    attr = composite_find_attribute(gridinfo, "visible_y");
-    if (attr != NULL)
-      data->grid.visible_y = data_int(attribute_first_data(attr));
-
-    data->grid.colour = prefs.new_diagram.grid_color;
-    attr = composite_find_attribute(diagramdata, "color");
-    if (attr != NULL)
-      data_color(attribute_first_data(attr), &data->grid.colour);
-
-  }
-  attr = composite_find_attribute(diagramdata, "guides");
-  if (attr != NULL) {
-    guint i;
-    DataNode guide;
-
-    guideinfo = attribute_first_data(attr);
-
-    attr = composite_find_attribute(guideinfo, "hguides");
+  if (diagram) {
+    attr = composite_find_attribute(diagramdata, "grid");
     if (attr != NULL) {
-      data->guides.nhguides = attribute_num_data(attr);
-      g_free(data->guides.hguides);
-      data->guides.hguides = g_new(real, data->guides.nhguides);
-    
-      guide = attribute_first_data(attr);
-      for (i = 0; i < data->guides.nhguides; i++, guide = data_next(guide))
-	data->guides.hguides[i] = data_real(guide);
+      gridinfo = attribute_first_data(attr);
+
+      attr = composite_find_attribute(gridinfo, "width_x");
+      if (attr != NULL)
+        diagram->grid.width_x = data_real(attribute_first_data(attr));
+      attr = composite_find_attribute(gridinfo, "width_y");
+      if (attr != NULL)
+        diagram->grid.width_y = data_real(attribute_first_data(attr));
+      attr = composite_find_attribute(gridinfo, "visible_x");
+      if (attr != NULL)
+        diagram->grid.visible_x = data_int(attribute_first_data(attr));
+      attr = composite_find_attribute(gridinfo, "visible_y");
+      if (attr != NULL)
+        diagram->grid.visible_y = data_int(attribute_first_data(attr));
+
+      diagram->grid.colour = prefs.new_diagram.grid_color;
+      attr = composite_find_attribute(diagramdata, "color");
+      if (attr != NULL)
+        data_color(attribute_first_data(attr), &diagram->grid.colour);
     }
-    attr = composite_find_attribute(guideinfo, "vguides");
+  }
+  if (diagram) {
+    attr = composite_find_attribute(diagramdata, "guides");
     if (attr != NULL) {
-      data->guides.nvguides = attribute_num_data(attr);
-      g_free(data->guides.vguides);
-      data->guides.vguides = g_new(real, data->guides.nvguides);
+      guint i;
+      DataNode guide;
+
+      guideinfo = attribute_first_data(attr);
+
+      attr = composite_find_attribute(guideinfo, "hguides");
+      if (attr != NULL) {
+        diagram->guides.nhguides = attribute_num_data(attr);
+        g_free(diagram->guides.hguides);
+        diagram->guides.hguides = g_new(real, diagram->guides.nhguides);
     
-      guide = attribute_first_data(attr);
-      for (i = 0; i < data->guides.nvguides; i++, guide = data_next(guide))
-	data->guides.vguides[i] = data_real(guide);
+        guide = attribute_first_data(attr);
+        for (i = 0; i < diagram->guides.nhguides; i++, guide = data_next(guide))
+	  diagram->guides.hguides[i] = data_real(guide);
+      }
+      attr = composite_find_attribute(guideinfo, "vguides");
+      if (attr != NULL) {
+        diagram->guides.nvguides = attribute_num_data(attr);
+        g_free(diagram->guides.vguides);
+        diagram->guides.vguides = g_new(real, diagram->guides.nvguides);
+    
+        guide = attribute_first_data(attr);
+        for (i = 0; i < diagram->guides.nvguides; i++, guide = data_next(guide))
+	  diagram->guides.vguides[i] = data_real(guide);
+      }
     }
   }
 
@@ -530,8 +576,8 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
 
     /* Read in all objects: */
     
-    list = read_objects(layer_node, layer, objects_hash, filename, NULL);
-    layer->objects = list;
+    list = read_objects(layer_node, objects_hash, filename, NULL);
+    layer_add_objects (layer, list);
     read_connections( list, layer_node, objects_hash);
 
     data_add_layer(data, layer);
@@ -550,7 +596,7 @@ diagram_data_load(const char *filename, DiagramData *data, void* user_data)
   if (data->layers->len < 1) {
     message_error (_("Error loading diagram:\n%s.\n"
                      "A valid Dia file defines at least one layer."),
-		     filename);
+		     dia_message_filename(filename));
     return FALSE;
   }
 
@@ -568,7 +614,7 @@ write_objects(GList *objects, xmlNodePtr objects_node,
 
   list = objects;
   while (list != NULL) {
-    Object *obj = (Object *) list->data;
+    DiaObject *obj = (DiaObject *) list->data;
 
     if (g_hash_table_lookup(objects_hash, obj))
     {
@@ -576,7 +622,7 @@ write_objects(GList *objects, xmlNodePtr objects_node,
       continue;
     }
 
-    if IS_GROUP(obj) {
+    if (IS_GROUP(obj) && group_objects(obj) != NULL) {
       group_node = xmlNewChild(objects_node, NULL, "group", NULL);
       write_objects(group_objects(obj), group_node,
 		    objects_hash, obj_nr, filename);
@@ -631,7 +677,7 @@ write_connections(GList *objects, xmlNodePtr layer_node,
   list = objects;
   obj_node = layer_node->xmlChildrenNode;
   while ((list != NULL) && (obj_node != NULL)) {
-    Object *obj = (Object *) list->data;
+    DiaObject *obj = (DiaObject *) list->data;
 
     while (obj_node && xmlIsBlankNode(obj_node)) obj_node = obj_node->next;
     if (!obj_node) break;
@@ -649,7 +695,7 @@ write_connections(GList *objects, xmlNodePtr layer_node,
 	con_point = handle->connected_to;
 	
 	if ( con_point != NULL ) {
-	  Object *other_obj;
+	  DiaObject *other_obj;
 	  int con_point_nr;
 	  
 	  other_obj = con_point->object;
@@ -677,16 +723,18 @@ write_connections(GList *objects, xmlNodePtr layer_node,
 
 	  xmlSetProp(connection, "to", buffer);
 	  /* to what connection_point on that object */
-	  g_snprintf(buffer, 30, "%d", con_point_nr);
+	  if (other_obj->connections[con_point_nr]->name != NULL) {
+	    g_snprintf(buffer, 30, "%s", other_obj->connections[con_point_nr]->name);
+	  } else {
+	    g_snprintf(buffer, 30, "%d", con_point_nr);
+	  }
 	  xmlSetProp(connection, "connection", buffer);
-
-
 	}
       }
     }
 
     if (obj->parent) {
-      Object *other_obj = obj->parent;
+      DiaObject *other_obj = obj->parent;
       xmlNodePtr parent;
       g_snprintf(buffer, 30, "O%d",
 		 GPOINTER_TO_INT(g_hash_table_lookup(objects_hash, other_obj)));
@@ -715,13 +763,14 @@ diagram_data_write_doc(DiagramData *data, const char *filename)
   Layer *layer;
   AttributeNode attr;
   xmlNs *name_space;
+  Diagram *diagram = DIA_IS_DIAGRAM (data) ? DIA_DIAGRAM (data) : NULL;
 
   doc = xmlNewDoc("1.0");
   doc->encoding = xmlStrdup("UTF-8");
   doc->xmlRootNode = xmlNewDocNode(doc, NULL, "diagram", NULL);
 
   name_space = xmlNewNs(doc->xmlRootNode, 
-                        "http://www.lysator.liu.se/~alla/dia/",
+                        DIA_XML_NAME_SPACE_BASE,
 			"dia");
   xmlSetNs(doc->xmlRootNode, name_space);
 
@@ -730,9 +779,10 @@ diagram_data_write_doc(DiagramData *data, const char *filename)
   attr = new_attribute((ObjectNode)tree, "background");
   data_add_color(attr, &data->bg_color);
 
-  attr = new_attribute((ObjectNode)tree, "pagebreak");
-  data_add_color(attr, &data->pagebreak_color);
-
+  if (diagram) {
+    attr = new_attribute((ObjectNode)tree, "pagebreak");
+    data_add_color(attr, &diagram->pagebreak_color);
+  }
   attr = new_attribute((ObjectNode)tree, "paper");
   pageinfo = data_add_composite(attr, "paper");
   data_add_string(composite_add_attribute(pageinfo, "name"),
@@ -758,29 +808,30 @@ diagram_data_write_doc(DiagramData *data, const char *filename)
 		 data->paper.fitheight);
   }
 
-  attr = new_attribute((ObjectNode)tree, "grid");
-  gridinfo = data_add_composite(attr, "grid");
-  data_add_real(composite_add_attribute(gridinfo, "width_x"),
-		data->grid.width_x);
-  data_add_real(composite_add_attribute(gridinfo, "width_y"),
-		data->grid.width_y);
-  data_add_int(composite_add_attribute(gridinfo, "visible_x"),
-	       data->grid.visible_x);
-  data_add_int(composite_add_attribute(gridinfo, "visible_y"),
-	       data->grid.visible_y);
-  attr = new_attribute((ObjectNode)tree, "color");
-  data_add_composite(gridinfo, "color");
-  data_add_color(attr, &data->grid.colour);
+  if (diagram) {
+    attr = new_attribute((ObjectNode)tree, "grid");
+    gridinfo = data_add_composite(attr, "grid");
+    data_add_real(composite_add_attribute(gridinfo, "width_x"),
+		  diagram->grid.width_x);
+    data_add_real(composite_add_attribute(gridinfo, "width_y"),
+		  diagram->grid.width_y);
+    data_add_int(composite_add_attribute(gridinfo, "visible_x"),
+	         diagram->grid.visible_x);
+    data_add_int(composite_add_attribute(gridinfo, "visible_y"),
+	         diagram->grid.visible_y);
+    attr = new_attribute((ObjectNode)tree, "color");
+    data_add_composite(gridinfo, "color");
+    data_add_color(attr, &diagram->grid.colour);
   
-
-  attr = new_attribute((ObjectNode)tree, "guides");
-  guideinfo = data_add_composite(attr, "guides");
-  attr = composite_add_attribute(guideinfo, "hguides");
-  for (i = 0; i < data->guides.nhguides; i++)
-    data_add_real(attr, data->guides.hguides[i]);
-  attr = composite_add_attribute(guideinfo, "vguides");
-  for (i = 0; i < data->guides.nvguides; i++)
-    data_add_real(attr, data->guides.vguides[i]);
+    attr = new_attribute((ObjectNode)tree, "guides");
+    guideinfo = data_add_composite(attr, "guides");
+    attr = composite_add_attribute(guideinfo, "hguides");
+    for (i = 0; i < diagram->guides.nhguides; i++)
+      data_add_real(attr, diagram->guides.hguides[i]);
+    attr = composite_add_attribute(guideinfo, "vguides");
+    for (i = 0; i < diagram->guides.nvguides; i++)
+    data_add_real(attr, diagram->guides.vguides[i]);
+  }
 
   objects_hash = g_hash_table_new(g_direct_hash, g_direct_equal);
 
@@ -864,7 +915,8 @@ diagram_data_save(DiagramData *data, const char *filename)
   /* Now write the data in the temporary file name. */
 
   if (file==NULL) {
-    message_error(_("Can't open output file %s: %s\n"), tmpname, strerror(errno));
+    message_error(_("Can't open output file %s: %s\n"), 
+		  dia_message_filename(tmpname), strerror(errno));
     return FALSE;
   }
   fclose(file);
@@ -897,11 +949,13 @@ diagram_save(Diagram *dia, const char *filename)
   gboolean res = diagram_data_save(dia->data, filename);
 
   if (!res) {
-    message_error(_("Failed to save file '%s'.\n"), filename);    
+    message_error(_("Failed to save file '%s'.\n"), 
+		  dia_message_filename(filename));
     return res;
   }
 
   dia->unsaved = FALSE;
+  undo_mark_save(dia->undo);
   diagram_set_modified (dia, FALSE);
 
   diagram_cleanup_autosave(dia);
@@ -917,9 +971,11 @@ diagram_cleanup_autosave(Diagram *dia)
   struct stat statbuf;
 
   savefile = dia->autosavefilename;
-
   if (savefile == NULL) return;
-
+#ifdef TRACES
+  g_print("Cleaning up autosave %s for %s\n", 
+          savefile, dia->filename ? dia->filename : "<no name>");
+#endif
   if (stat(savefile, &statbuf) == 0) { /* Success */
     unlink(savefile);
   }
@@ -942,7 +998,7 @@ diagram_autosave(Diagram *dia)
   while (diagrams != NULL) {
     diagram = (Diagram *)diagrams->data;
     if (diagram == dia &&
-	diagram->modified && 
+	diagram_is_modified(diagram) && 
 	!diagram->autosaved) {
       save_filename = g_strdup_printf("%s.autosave", dia->filename);
 
@@ -967,12 +1023,12 @@ export_native(DiagramData *data, const gchar *filename,
 
 static const gchar *extensions[] = { "dia", NULL };
 DiaExportFilter dia_export_filter = {
-  N_("Native Dia Diagram"),
+  N_("Dia Diagram File"),
   extensions,
   export_native
 };
 DiaImportFilter dia_import_filter = {
-  N_("Native Dia Diagram"),
+  N_("Dia Diagram File"),
   extensions,
   diagram_data_load
 };
