@@ -33,13 +33,15 @@ select_all_callback(gpointer data, guint action, GtkWidget *widget)
 {
   Diagram *dia;
   GList *objects;
+  DDisplay * ddisp = ddisplay_active();
 
-  dia = ddisplay_active()->diagram;
+  if (!ddisp) return;
+  dia = ddisp->diagram;
 
   objects = dia->data->active_layer->objects;
 
   while (objects != NULL) {
-    Object *obj = (Object *)objects->data;
+    DiaObject *obj = (DiaObject *)objects->data;
     
     if (!diagram_is_selected(dia, obj))
       diagram_select(dia, obj);
@@ -47,7 +49,7 @@ select_all_callback(gpointer data, guint action, GtkWidget *widget)
     objects = g_list_next(objects);
   }
 
-  diagram_update_menu_sensitivity(dia);
+  ddisplay_do_update_menu_sensitivity(ddisp);
   object_add_updates_list(dia->data->selected, dia);
   diagram_flush(dia);
 }
@@ -55,11 +57,15 @@ select_all_callback(gpointer data, guint action, GtkWidget *widget)
 void
 select_none_callback(gpointer data, guint action, GtkWidget *widget)
 {
-  Diagram * dia = ddisplay_active()->diagram;
+  Diagram * dia;
+  DDisplay * ddisp = ddisplay_active();
+
+  if (!ddisp) return;
+  dia = ddisp->diagram;
 
   diagram_remove_all_selected(dia, TRUE);
 
-  diagram_update_menu_sensitivity(dia);
+  ddisplay_do_update_menu_sensitivity(ddisp);
   object_add_updates_list(dia->data->selected, dia);
   diagram_flush(dia);
 }
@@ -69,13 +75,15 @@ select_invert_callback(gpointer data, guint action, GtkWidget *widget)
 {
   Diagram *dia;
   GList *tmp;
+  DDisplay * ddisp = ddisplay_active();
 
-  dia = ddisplay_active()->diagram;
+  if (!ddisp) return;
+  dia = ddisp->diagram;
 
   tmp = dia->data->active_layer->objects;
 
   for (; tmp != NULL; tmp = g_list_next(tmp)) {
-    Object *obj = (Object *)tmp->data;
+    DiaObject *obj = (DiaObject *)tmp->data;
     
     if (!diagram_is_selected(dia, obj))
       diagram_select(dia, obj);
@@ -83,28 +91,36 @@ select_invert_callback(gpointer data, guint action, GtkWidget *widget)
       diagram_unselect_object(dia, obj);
   }
 
-  diagram_update_menu_sensitivity(dia);
+  ddisplay_do_update_menu_sensitivity(ddisp);
   object_add_updates_list(dia->data->selected, dia);
   diagram_flush(dia);
   
 }
 
+/** Select objects that are directly connected to the currently selected
+ * objects, but only in the active layer.
+ */
 void
 select_connected_callback(gpointer data, guint action, GtkWidget *widget)
 {
-  Diagram *dia = ddisplay_active()->diagram;
+  Diagram *dia;
+  DDisplay * ddisp = ddisplay_active();
   GList *objects, *tmp;
 
+  if (!ddisp) return;
+  dia = ddisp->diagram;
+
   objects = dia->data->selected;
 
   for (tmp = objects; tmp != NULL; tmp = g_list_next(tmp)) {
-    Object *obj = (Object *)tmp->data;
+    DiaObject *obj = (DiaObject *)tmp->data;
     int i;
 
     for (i = 0; i < obj->num_handles; i++) {
       Handle *handle = obj->handles[i];
       
-      if (handle->connected_to != NULL) {
+      if (handle->connected_to != NULL &&
+	handle->connected_to->object->parent_layer == dia->data->active_layer) {
 	if (!diagram_is_selected(dia, handle->connected_to->object)) {
           diagram_select(dia, handle->connected_to->object);
 	}
@@ -113,7 +129,7 @@ select_connected_callback(gpointer data, guint action, GtkWidget *widget)
   }
 
   for (tmp = objects; tmp != NULL; tmp = tmp->next) {
-    Object *obj = (Object *)tmp->data;
+    DiaObject *obj = (DiaObject *)tmp->data;
     int i;
 
     for (i = 0; i < obj->num_connections; i++) {
@@ -121,22 +137,23 @@ select_connected_callback(gpointer data, guint action, GtkWidget *widget)
       GList *conns = connection->connected;
 
       for (; conns != NULL; conns = g_list_next(conns)) {
-        Object *obj2 = (Object *)conns->data;
+        DiaObject *obj2 = (DiaObject *)conns->data;
 
-	if (!diagram_is_selected(dia, obj2)) {
+	if (obj2->parent_layer == dia->data->active_layer
+	    && !diagram_is_selected(dia, obj2)) {
           diagram_select(dia, obj2);
 	}
       }      
     }
   }
 
-  diagram_update_menu_sensitivity(dia);
+  ddisplay_do_update_menu_sensitivity(ddisp);
   object_add_updates_list(dia->data->selected, dia);
   diagram_flush(dia);
 }
 
 static void
-select_transitively(Diagram *dia, Object *obj)
+select_transitively(Diagram *dia, DiaObject *obj)
 {
   int i;
   /* Do breadth-first to avoid overly large stack */
@@ -145,7 +162,8 @@ select_transitively(Diagram *dia, Object *obj)
   for (i = 0; i < obj->num_handles; i++) {
     Handle *handle = obj->handles[i];
     
-    if (handle->connected_to != NULL) {
+    if (handle->connected_to != NULL &&
+	handle->connected_to->object->parent_layer == dia->data->active_layer) {
       if (!diagram_is_selected(dia, handle->connected_to->object)) {
 	diagram_select(dia, handle->connected_to->object);
 	newly_selected = g_list_prepend(newly_selected, handle->connected_to->object);
@@ -158,9 +176,10 @@ select_transitively(Diagram *dia, Object *obj)
     GList *conns = connection->connected;
 
     for (; conns != NULL; conns = g_list_next(conns)) {
-      Object *obj2 = (Object *)conns->data;
+      DiaObject *obj2 = (DiaObject *)conns->data;
 
-      if (!diagram_is_selected(dia, obj2)) {
+      if (obj2->parent_layer == dia->data->active_layer
+	  && !diagram_is_selected(dia, obj2)) {
 	diagram_select(dia, obj2);
 	newly_selected = g_list_prepend(newly_selected, obj2);
       }
@@ -168,24 +187,31 @@ select_transitively(Diagram *dia, Object *obj)
   }
   
   while (newly_selected != NULL) {
-    select_transitively(dia, (Object *)newly_selected->data);
+    select_transitively(dia, (DiaObject *)newly_selected->data);
     newly_selected = g_list_next(newly_selected);
   }
 }
 
+/** Select objects that are in any way connected with a currently selected
+ * object, but only in the active layer.
+ */
 void
 select_transitive_callback(gpointer data, guint action, GtkWidget *widget)
 {
-  Diagram *dia = ddisplay_active()->diagram;
+  DDisplay *ddisp = ddisplay_active();
+  Diagram *dia;
   GList *objects, *tmp;
 
+  if (!ddisp) return;
+  dia = ddisp->diagram;
+
   objects = dia->data->selected;
 
   for (tmp = objects; tmp != NULL; tmp = g_list_next(tmp)) {
-    select_transitively(dia, (Object *)tmp->data);
+    select_transitively(dia, (DiaObject *)tmp->data);
   }
 
-  diagram_update_menu_sensitivity(dia);
+  ddisplay_do_update_menu_sensitivity(ddisp);
   object_add_updates_list(dia->data->selected, dia);
   diagram_flush(dia);
 }
@@ -195,21 +221,23 @@ select_same_type_callback(gpointer data, guint action, GtkWidget *widget)
 {
   /* For now, do a brute force version:  Check vs. all earlier selected.
      Later, we should really sort the selecteds first to avoid n^2 */
+  DDisplay *ddisp = ddisplay_active();
   Diagram *dia;
   GList *objects, *tmp, *tmp2;
 
-  dia = ddisplay_active()->diagram;
+  if (!ddisp) return;
+  dia = ddisp->diagram;
 
   tmp = dia->data->active_layer->objects;
 
   objects = dia->data->selected;
 
   for (; tmp != NULL; tmp = g_list_next(tmp)) {
-    Object *obj = (Object *)tmp->data;
+    DiaObject *obj = (DiaObject *)tmp->data;
 
     if (!diagram_is_selected(dia, obj)) {
       for (tmp2 = objects; tmp2 != NULL; tmp2 = g_list_next(tmp2)) {
-        Object *obj2 = (Object *)tmp2->data;
+        DiaObject *obj2 = (DiaObject *)tmp2->data;
        
         if (obj->type == obj2->type) {
           diagram_select(dia, obj);
@@ -219,7 +247,7 @@ select_same_type_callback(gpointer data, guint action, GtkWidget *widget)
     }
   }
 
-  diagram_update_menu_sensitivity(dia);
+  ddisplay_do_update_menu_sensitivity(ddisp);
   object_add_updates_list(dia->data->selected, dia);
   diagram_flush(dia);
 }
@@ -233,3 +261,7 @@ select_style_callback(gpointer data, guint action, GtkWidget *widget)
   selection_style = action;
 }
 
+
+
+
+
