@@ -28,6 +28,7 @@
 #include "font.h"
 #include "intl.h"
 #include "dia_image.h"
+#include "object.h"
 
 #ifdef HAVE_LIBART
 
@@ -38,8 +39,10 @@
 /* ugly namespace clashes */
 #define Rectangle Win32Rectangle
 #include <pango/pango.h>
-#include <pango/pangowin32.h>
 #undef Rectangle
+#else
+#include <pango/pango.h>
+#include <gdk/gdk.h>
 #endif
 
 #include <libart_lgpl/art_point.h>
@@ -69,15 +72,22 @@ color_to_abgr(Color *col)
 }
 
 static inline guint32 
-color_to_rgba(Color *col)
+color_to_rgba(DiaLibartRenderer *renderer, Color *col)
 {
   int rgba;
 
-  rgba = 0xFF;
-  rgba |= (guint)(0xFF*col->red) << 24;
-  rgba |= (guint)(0xFF*col->green) << 16;
-  rgba |= (guint)(0xFF*col->blue) << 8;
-  
+  if (renderer->highlight_color != NULL) {
+      rgba = 0xFF;
+      rgba |= (guint)(0xFF*renderer->highlight_color->red) << 24;
+      rgba |= (guint)(0xFF*renderer->highlight_color->green) << 16;
+      rgba |= (guint)(0xFF*renderer->highlight_color->blue) << 8;
+  } else {
+    rgba = 0xFF;
+    rgba |= (guint)(0xFF*col->red) << 24;
+    rgba |= (guint)(0xFF*col->green) << 16;
+    rgba |= (guint)(0xFF*col->blue) << 8;
+  }
+
   return rgba;
 }
 
@@ -100,7 +110,6 @@ get_height_pixels (DiaRenderer *self)
 static void
 begin_render(DiaRenderer *self)
 {
-  DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
 #ifdef HAVE_FREETYPE
   /* pango_ft2_get_context API docs :
    * ... Use of this function is discouraged, ...
@@ -110,17 +119,21 @@ begin_render(DiaRenderer *self)
   dia_font_push_context(pango_ft2_get_context(75, 75));
 # define FONT_SCALE (1.0)
 #elif defined G_OS_WIN32
-  dia_font_push_context(pango_win32_get_context());
-  /* we need to scale but can't as simple as with ft2 */
-# define FONT_SCALE (1.0 / 22.0)
+  dia_font_push_context(gdk_pango_context_get());
+  /* I shall never claim again to have understood Dias/Pangos font size
+   * relations ;). This was (1.0/22.0) but nowadays 1.0 seems to be
+   * fine with Pango/win32, too.  --hb
+   */
+# define FONT_SCALE (1.0)
+#else
+  dia_font_push_context (gdk_pango_context_get ());
+# define FONT_SCALE (0.8)
 #endif
 }
 
 static void
 end_render(DiaRenderer *self)
 {
-  DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
-
   dia_font_pop_context();
 }
 
@@ -131,9 +144,13 @@ set_linewidth(DiaRenderer *self, real linewidth)
 {  /* 0 == hairline **/
   DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
 
+  if (renderer->highlight_color != NULL) {
+    /* 6 pixels wide -> 3 pixels beyond normal obj */
+    real border = dia_untransform_length(renderer->transform, 6);
+    linewidth += border;
+  } 
   renderer->line_width =
     dia_transform_length(renderer->transform, linewidth);
-
   if (renderer->line_width<=0.5)
     renderer->line_width = 0.5; /* Minimum 0.5 pixel. */
 }
@@ -143,16 +160,21 @@ set_linecaps(DiaRenderer *self, LineCaps mode)
 {
   DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
 
-  switch(mode) {
-  case LINECAPS_BUTT:
-    renderer->cap_style = ART_PATH_STROKE_CAP_BUTT;
-    break;
-  case LINECAPS_ROUND:
+  if (renderer->highlight_color != NULL) {
+    /* Can't tell that this does anything:( -LC */
     renderer->cap_style = ART_PATH_STROKE_CAP_ROUND;
-    break;
-  case LINECAPS_PROJECTING:
-    renderer->cap_style = ART_PATH_STROKE_CAP_SQUARE;
-    break;
+  } else {
+    switch(mode) {
+    case LINECAPS_BUTT:
+      renderer->cap_style = ART_PATH_STROKE_CAP_BUTT;
+      break;
+    case LINECAPS_ROUND:
+      renderer->cap_style = ART_PATH_STROKE_CAP_ROUND;
+      break;
+    case LINECAPS_PROJECTING:
+      renderer->cap_style = ART_PATH_STROKE_CAP_SQUARE;
+      break;
+    }
   }
 }
 
@@ -161,16 +183,21 @@ set_linejoin(DiaRenderer *self, LineJoin mode)
 {
   DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
 
-  switch(mode) {
-  case LINEJOIN_MITER:
-    renderer->cap_style = ART_PATH_STROKE_JOIN_MITER;
-    break;
-  case LINEJOIN_ROUND:
-    renderer->cap_style = ART_PATH_STROKE_JOIN_ROUND;
-    break;
-  case LINEJOIN_BEVEL:
-    renderer->cap_style = ART_PATH_STROKE_JOIN_BEVEL;
-    break;
+  if (renderer->highlight_color != NULL) {
+    /* Can't tell that this does anything:( -LC */
+    renderer->join_style = ART_PATH_STROKE_JOIN_ROUND;
+  } else {
+    switch(mode) {
+    case LINEJOIN_MITER:
+      renderer->join_style = ART_PATH_STROKE_JOIN_MITER;
+      break;
+    case LINEJOIN_ROUND:
+      renderer->join_style = ART_PATH_STROKE_JOIN_ROUND;
+      break;
+    case LINEJOIN_BEVEL:
+      renderer->join_style = ART_PATH_STROKE_JOIN_BEVEL;
+      break;
+    }
   }
 }
 
@@ -259,8 +286,6 @@ set_dashlength(DiaRenderer *self, real length)
 static void
 set_fillstyle(DiaRenderer *self, FillStyle mode)
 {
-  DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
-
   switch(mode) {
   case FILLSTYLE_SOLID:
     break;
@@ -273,11 +298,11 @@ static void
 set_font(DiaRenderer *self, DiaFont *font, real height)
 {
   self->font_height = height * FONT_SCALE;
-  //    ddisplay_transform_length(renderer->ddisp, height);
 
+  dia_font_ref(font);
   if (self->font)
     dia_font_unref(self->font);
-  self->font = dia_font_ref(font);
+  self->font = font;
 }
 
 static void
@@ -291,7 +316,7 @@ draw_line(DiaRenderer *self,
   guint32 rgba;
   double x,y;
 
-  rgba = color_to_rgba(line_color);
+  rgba = color_to_rgba(renderer, line_color);
   
   vpath = art_new (ArtVpath, 3);
   
@@ -347,7 +372,7 @@ draw_polyline(DiaRenderer *self,
   double x,y;
   int i;
 
-  rgba = color_to_rgba(line_color);
+  rgba = color_to_rgba(renderer, line_color);
   
   vpath = art_new (ArtVpath, num_points+1);
 
@@ -401,7 +426,7 @@ draw_polygon(DiaRenderer *self,
   double x,y;
   int i;
   
-  rgba = color_to_rgba(line_color);
+  rgba = color_to_rgba(renderer, line_color);
   
   vpath = art_new (ArtVpath, num_points+2);
 
@@ -461,7 +486,7 @@ fill_polygon(DiaRenderer *self,
   double x,y;
   int i;
   
-  rgba = color_to_rgba(color);
+  rgba = color_to_rgba(renderer, color);
   
   vpath = art_new (ArtVpath, num_points+2);
 
@@ -517,7 +542,7 @@ draw_rect(DiaRenderer *self,
   if ((left>right) || (top>bottom))
     return;
 
-  rgba = color_to_rgba(color);
+  rgba = color_to_rgba(renderer, color);
   
   vpath = art_new (ArtVpath, 6);
 
@@ -585,7 +610,7 @@ fill_rect(DiaRenderer *self,
   if ((left>right) || (top>bottom))
     return;
 
-  rgba = color_to_rgba(color);
+  rgba = color_to_rgba(renderer, color);
   
   vpath = art_new (ArtVpath, 6);
 
@@ -667,7 +692,7 @@ draw_arc(DiaRenderer *self,
   if (num_points<5) /* Don't be too coarse */
     num_points = 5;
 
-  rgba = color_to_rgba(line_color);
+  rgba = color_to_rgba(renderer, line_color);
   
   vpath = art_new (ArtVpath, num_points+1);
 
@@ -753,7 +778,7 @@ fill_arc(DiaRenderer *self,
   if (num_points<5) /* Don't be too coarse */
     num_points = 5;
 
-  rgba = color_to_rgba(color);
+  rgba = color_to_rgba(renderer, color);
   
   vpath = art_new (ArtVpath, num_points+2+1);
 
@@ -822,7 +847,7 @@ draw_bezier(DiaRenderer *self,
   double x,y;
   int i;
 
-  rgba = color_to_rgba(line_color);
+  rgba = color_to_rgba(renderer, line_color);
   
   bpath = art_new (ArtBpath, numpoints+1);
 
@@ -911,7 +936,7 @@ fill_bezier(DiaRenderer *self,
   double x,y;
   int i;
 
-  rgba = color_to_rgba(color);
+  rgba = color_to_rgba(renderer, color);
   
   bpath = art_new (ArtBpath, numpoints+1);
 
@@ -976,6 +1001,66 @@ fill_bezier(DiaRenderer *self,
 }
 
 
+/* Draw a highlighted version of a string.
+ */
+static void
+draw_highlighted_string(DiaLibartRenderer *renderer,
+			PangoLayout *layout,
+			real x, real y,
+			guint32 rgba)
+{
+  ArtVpath *vpath;
+  ArtSVP *svp;
+  int width, height;
+  double top, bottom, left, right;
+    
+  pango_layout_get_pixel_size(layout, &width, &height);
+  dia_transform_coords_double(renderer->transform,
+                              x, y, &left, &top);
+  left -= 3;
+  right = left+width+6;
+  bottom = top+height;
+  
+  if ((left>right) || (top>bottom))
+    return;
+
+  vpath = art_new (ArtVpath, 6);
+
+  vpath[0].code = ART_MOVETO;
+  vpath[0].x = left;
+  vpath[0].y = top;
+  vpath[1].code = ART_LINETO;
+  vpath[1].x = right;
+  vpath[1].y = top;
+  vpath[2].code = ART_LINETO;
+  vpath[2].x = right;
+  vpath[2].y = bottom;
+  vpath[3].code = ART_LINETO;
+  vpath[3].x = left;
+  vpath[3].y = bottom;
+  vpath[4].code = ART_LINETO;
+  vpath[4].x = left;
+  vpath[4].y = top;
+  vpath[5].code = ART_END;
+  vpath[5].x = 0;
+  vpath[5].y = 0;
+  
+  svp = art_svp_from_vpath (vpath);
+  
+  art_free( vpath );
+  
+  art_rgb_svp_alpha (svp,
+		     0, 0, 
+		     renderer->pixel_width,
+		     renderer->pixel_height,
+		     rgba,
+		     renderer->rgb_buffer, renderer->pixel_width*3,
+		     NULL);
+
+  art_svp_free( svp );
+}
+
+
 static void
 draw_string (DiaRenderer *self,
 	     const gchar *text,
@@ -990,17 +1075,15 @@ draw_string (DiaRenderer *self,
   PangoLayout* layout;
   int width, height;
   int i, j;
-  int iwidth;
-  int len;
   double affine[6], tmpaffine[6];
-  double xpos, ypos;
   guint32 rgba;
   int rowstride;
-  double zoom;
   double font_height;
 
   point_copy(&start_pos,pos);
 
+  rgba = color_to_rgba(renderer, color);
+
   /* Something's screwed up with the zooming for fonts.
      It's like it zooms double.  Yet removing some of the zooms
      don't seem to work, either.
@@ -1037,6 +1120,13 @@ draw_string (DiaRenderer *self,
   layout = dia_font_scaled_build_layout(
               text, self->font, self->font_height,
               dia_transform_length(renderer->transform, 1.0));
+
+  if (renderer->highlight_color != NULL) {
+    draw_highlighted_string(renderer, layout, start_pos.x, start_pos.y, rgba);
+    g_object_unref(G_OBJECT(layout));
+    return;
+  }
+
   /*
   ddisp->zoom_factor = old_zoom;
   */
@@ -1051,15 +1141,9 @@ draw_string (DiaRenderer *self,
  {
    FT_Bitmap ftbitmap;
    guint8 *graybitmap;
-   FT_Face *face;
-   PangoFont *font;
 
    rowstride = 32*((width+31)/31);
    
-   font = pango_context_load_font(pango_ft2_get_context(75, 75),
-				  self->font->pfd);
-   face = pango_ft2_font_get_face(font);
-
    graybitmap = (guint8*)g_new0(guint8, height*rowstride);
 
    ftbitmap.rows = height;
@@ -1079,34 +1163,9 @@ draw_string (DiaRenderer *self,
        bitmap[DEPTH*(i*rowstride+j)+2] = color->blue*255;
        bitmap[DEPTH*(i*rowstride+j)+3] = graybitmap[i*rowstride+j];
      }
-     //     bitmap[4*i+3] = graybitmap[i];
    }
    g_free(graybitmap);
  }
-#elif defined G_OS_WIN32 && defined PANGO_WIN32_FUTURE
- /* duplicating from above, please let extending Pango be allowed, bug 94791 */
- {
-   PangoBitmap graybitmap;
-   rowstride = 32*((width+31)/31);
-   
-   graybitmap.rows = height;
-   graybitmap.width = width;
-   graybitmap.pitch = rowstride;
-   graybitmap.buffer = (guint8*)g_new0(guint8, height*rowstride);
-   graybitmap.num_grays = 256;
-   g_print("Rendering %s onto bitmap of size %dx%d row %d\n", text, width, height, rowstride);
-   pango_win32_render_layout_to_bitmap(&graybitmap, layout, 0, 0);
-   bitmap = (guint8*)g_new0(guint8, height*rowstride*DEPTH);
-   for (i = 0; i < height; i++) {
-     for (j = 0; j < width; j++) {
-       bitmap[DEPTH*(i*rowstride+j)] = color->red*255;
-       bitmap[DEPTH*(i*rowstride+j)+1] = color->green*255;
-       bitmap[DEPTH*(i*rowstride+j)+2] = color->blue*255;
-       bitmap[DEPTH*(i*rowstride+j)+3] = graybitmap.buffer[i*rowstride+j];
-     }
-   }
-   g_free(graybitmap.buffer);
- }
 #else
  /* gdk does not like 0x0 sized (it does not make much sense with the others as well,
   * but they don't complain ;) */
@@ -1117,13 +1176,13 @@ draw_string (DiaRenderer *self,
    GdkImage  *image;
 
    rowstride = 32*((width+31)/31);
-#if 0 /* with 8 bit pixmap we would probably need to set the whole gray palette */
-   gdk_gc_set_foreground (gc, &color_gdk_white);
+#if 1 /* with 8 bit pixmap we would probably need to set the whole gray palette */
+   gdk_gc_set_foreground (gc, &color_gdk_black);
    gdk_gc_set_background (gc, &color_gdk_white);
    gdk_draw_rectangle (GDK_DRAWABLE (pixmap), gc, TRUE, 0, 0, width, height); 
 #endif
-   gdk_gc_set_foreground (gc, &color_gdk_black);
-   gdk_gc_set_background (gc, &color_gdk_white);
+   gdk_gc_set_foreground (gc, &color_gdk_white);
+   gdk_gc_set_background (gc, &color_gdk_black);
    gdk_draw_layout (GDK_DRAWABLE (pixmap), gc, 0, 0, layout);
    image = gdk_drawable_get_image (GDK_DRAWABLE (pixmap), 0, 0, width, height);
    g_object_unref (G_OBJECT (gc));
@@ -1134,9 +1193,10 @@ draw_string (DiaRenderer *self,
        bitmap[DEPTH*(i*rowstride+j)] = color->red*255;
        bitmap[DEPTH*(i*rowstride+j)+1] = color->green*255;
        bitmap[DEPTH*(i*rowstride+j)+2] = color->blue*255;
-       bitmap[DEPTH*(i*rowstride+j)+3] = gdk_image_get_pixel (image, j, i) ? 255 : 0;
+       bitmap[DEPTH*(i*rowstride+j)+3] = gdk_image_get_pixel (image, j, i) & 0xFF;
      }
    }
+   g_object_unref (G_OBJECT (image));
  }
 #endif
 
@@ -1144,8 +1204,6 @@ draw_string (DiaRenderer *self,
   
   g_object_unref(G_OBJECT(layout));
 
-  rgba = color_to_rgba(color);
-
   art_affine_identity(affine);
   art_affine_translate(tmpaffine, x, y);
   art_affine_multiply(affine, affine, tmpaffine);
@@ -1160,7 +1218,6 @@ draw_string (DiaRenderer *self,
 		    width,
 		    height,
 		    rowstride*DEPTH,
-		    //rgba,
 		    affine,
 		    ART_FILTER_NEAREST, NULL);
 
@@ -1179,7 +1236,7 @@ get_text_width(DiaRenderer *object,
   if (length != strlen(text)) {
     char *othertx;
     int ulen;
-    /* A couple UTF8-chars: æblegrød Š Ť Ž ę ć ń уфхцНОПРЄ є Ґ Њ Ћ Џ */
+    /* A couple UTF8-chars: æblegrød Š Ť Ž ę ć ń уфхц?ОПРЄ є ? Њ Ћ ? */
     ulen = g_utf8_offset_to_pointer(text, length)-text;
     if (!g_utf8_validate(text, ulen, NULL)) {
       g_warning ("Text at char %d not valid\n", length);
@@ -1208,77 +1265,98 @@ draw_image(DiaRenderer *self,
 	   DiaImage image)
 {
   DiaLibartRenderer *renderer = DIA_LIBART_RENDERER (self);
-  double real_width, real_height;
-  double x,y;
-  int src_width, src_height;
-  guint8 *img_data;
-  guint8 *img_mask;
-  double affine[6];
-  int i,j;
-  int rowstride;
 
   /* Todo: Handle some kind of clipping! */
   
-  real_width = dia_transform_length(renderer->transform, width);
-  real_height = dia_transform_length(renderer->transform, height);
-  dia_transform_coords_double(renderer->transform, 
-                              point->x, point->y, &x, &y);
+  if (renderer->highlight_color != NULL) {
+    Point lr;
+    DiaRendererClass *self_class = DIA_RENDERER_GET_CLASS (self);
 
-  src_width = dia_image_width(image);
-  src_height = dia_image_height(image);
-  rowstride = dia_image_rowstride(image);
-
-  affine[0] = real_width/(double)src_width;
-  affine[1] = 0;
-  affine[2] = 0;
-  affine[3] = real_height/(double)src_height;
-  affine[4] = x;
-  affine[5] = y;
-
-  img_data = dia_image_rgba_data(image);
-
-  if (img_data != NULL) {
-    /* If there is an alpha channel, we can use it directly. */
-    art_rgb_rgba_affine(renderer->rgb_buffer,
-			0, 0,
-			renderer->pixel_width,
-			renderer->pixel_height,
-			renderer->pixel_width*3,
-			img_data, src_width, src_height, 
-			rowstride,
-			affine, ART_FILTER_NEAREST, NULL);
-    /* Note that dia_image_rgba_data doesn't copy */
+    lr = *point;
+    lr.x += width;
+    lr.y += height;
+    self_class->fill_rect(self, point, &lr, renderer->highlight_color);
   } else {
-    img_data = dia_image_rgb_data(image);
+    double real_width, real_height;
+    double x,y;
+    int src_width, src_height;
+    double affine[6];
+    int rowstride;
+    real_width = dia_transform_length(renderer->transform, width);
+    real_height = dia_transform_length(renderer->transform, height);
+    dia_transform_coords_double(renderer->transform, 
+				point->x, point->y, &x, &y);
 
-    art_rgb_affine(renderer->rgb_buffer,
-		   0, 0,
-		   renderer->pixel_width,
-		   renderer->pixel_height,
-		   renderer->pixel_width*3,
-		   img_data, src_width, src_height, 
-		   rowstride,
-		   affine, ART_FILTER_NEAREST, NULL);
+    src_width = dia_image_width(image);
+    src_height = dia_image_height(image);
+    rowstride = dia_image_rowstride(image);
+
+    affine[0] = real_width/(double)src_width;
+    affine[1] = 0;
+    affine[2] = 0;
+    affine[3] = real_height/(double)src_height;
+    affine[4] = x;
+    affine[5] = y;
+
+    if (dia_image_rgba_data(image)) {
+      /* If there is an alpha channel, we can use it directly. */
+      const guint8 *img_data = dia_image_rgba_data(image);
+      art_rgb_rgba_affine(renderer->rgb_buffer,
+			  0, 0,
+			  renderer->pixel_width,
+			  renderer->pixel_height,
+			  renderer->pixel_width*3,
+			  img_data, src_width, src_height, 
+			  rowstride,
+			  affine, ART_FILTER_NEAREST, NULL);
+      /* Note that dia_image_rgba_data doesn't copy */
+    } else {
+      guint8 *img_data = dia_image_rgb_data(image);
+
+      art_rgb_affine(renderer->rgb_buffer,
+		     0, 0,
+		     renderer->pixel_width,
+		     renderer->pixel_height,
+		     renderer->pixel_width*3,
+		     img_data, src_width, src_height, 
+		     rowstride,
+		     affine, ART_FILTER_NEAREST, NULL);
     
-    g_free(img_data);
+      g_free(img_data);
+    }
   }
 }
 
 
+static void
+draw_object (DiaRenderer *renderer, DiaObject *object)
+{
+  if (renderer->is_interactive &&
+      object->highlight_color != NULL) {
+    DiaLibartRenderer *libart_rend = DIA_LIBART_RENDERER(renderer);
+    libart_rend->highlight_color = object->highlight_color;
+    object->ops->draw(object, renderer);
+    libart_rend->highlight_color = NULL;
+  }
+  object->ops->draw(object, renderer);
+}
+
 static void
 renderer_init (DiaLibartRenderer *renderer, gpointer g_class)
 {
   renderer->rgb_buffer = NULL;
 
   renderer->line_width = 1.0;
-  renderer->cap_style = GDK_CAP_BUTT;
-  renderer->join_style = GDK_JOIN_MITER;
+  renderer->cap_style = ART_PATH_STROKE_CAP_BUTT;
+  renderer->join_style = ART_PATH_STROKE_JOIN_MITER;
   
   renderer->saved_line_style = LINESTYLE_SOLID;
   renderer->dash_enabled = 0;
   renderer->dash_length = 10;
   renderer->dot_length = 1;
 
+  renderer->highlight_color = NULL;
+
   renderer->parent_instance.font = NULL;
 }
 
@@ -1293,9 +1371,6 @@ renderer_finalize (GObject *object)
   if (renderer->rgb_buffer != NULL)
     g_free(renderer->rgb_buffer);
 
-  if (renderer->parent_instance.font)
-    dia_font_unref(renderer->parent_instance.font);
-
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -1330,9 +1405,13 @@ dia_libart_renderer_get_type (void)
 static void
 dia_libart_renderer_class_init (DiaLibartRendererClass *klass)
 {
-  GObjectClass *object_class = G_OBJECT_CLASS (klass);
+  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   DiaRendererClass *renderer_class = DIA_RENDERER_CLASS (klass);
 
+  parent_class = g_type_class_peek_parent (klass);
+
+  gobject_class->finalize = renderer_finalize;
+	
   /* Here we set the functions that we define for this renderer. */
   renderer_class->get_width_pixels = get_width_pixels;
   renderer_class->get_height_pixels = get_height_pixels;
@@ -1370,6 +1449,8 @@ dia_libart_renderer_class_init (DiaLibartRendererClass *klass)
 
   renderer_class->draw_image = draw_image;
 
+  renderer_class->draw_object = draw_object;
+
   /* Interactive functions */
   renderer_class->get_text_width = get_text_width;
 }
